# encoding: utf-8

from collections import (
    namedtuple,
    OrderedDict,
)
from functools import cmp_to_key
from itertools import combinations

import six

from six.moves import (
    range,
    reduce,
)

from .pairs_storage import (
    PairsStorage,
    key,
)


class Item(object):

    @property
    def id(self):
        return self.__item_id

    @property
    def value(self):
        return self.__value

    def __init__(self, item_id, value):
        self.__item_id = item_id
        self.__value = value
        self.weights = []

    def __str__(self):
        return str(self.__dict__)


def get_max_combination_number(prameter_matrix, n):
    param_len_list = [len(value_list) for value_list in prameter_matrix]

    return sum([
        reduce(lambda x, y: x * y, z)
        for z in combinations(param_len_list, n)
    ])


def cmp_item(lhs, rhs):
    if lhs.weights == rhs.weights:
        return 0

    return -1 if lhs.weights < rhs.weights else 1


class AllPairs(object):

    @property
    def Pairs(self):
        return self.__Pairs

    def __init__(
            self, parameters, filter_func=lambda x: True,
            previously_tested=[[]], n=2):
        """
        TODO: check that input arrays are:
            - (optional) has no duplicated values inside single array / or compress such values
        """

        self.__validate_parameter(parameters)

        self.__is_ordered_dict_param = isinstance(parameters, OrderedDict)
        self.__param_name_list = self.__extract_param_name_list(parameters)
        self.__Pairs = namedtuple("Pairs", self.__param_name_list)

        self.__filter_func = filter_func
        self.__n = n
        self.__pairs = PairsStorage(n)

        value_matrix = self.__extract_value_matrix(parameters)
        self.__max_unique_pairs_expected = get_max_combination_number(
            value_matrix, n)
        self.__working_item_matrix = self.__get_working_item_matrix(
            value_matrix)

        for arr in previously_tested:
            if len(arr) == 0:
                continue

            if len(arr) != len(self.__working_item_matrix):
                raise RuntimeError(
                    "previously tested combination is not complete")

            if not self.__filter_func(arr):
                raise ValueError("invalid tested combination is provided")

            tested = []
            for i, val in enumerate(arr):
                idxs = [
                    Item(item.id, 0)
                    for item in self.__working_item_matrix[i]
                    if item.value == val
                ]

                if len(idxs) != 1:
                    raise ValueError(
                        "value from previously tested combination is not "
                        "found in the parameters or found more than "
                        "once")

                tested.append(idxs[0])

            self.__pairs.add_sequence(tested)

    def __iter__(self):
        return self

    def next(self):
        return self.__next__()

    def __next__(self):
        assert(len(self.__pairs) <= self.__max_unique_pairs_expected)

        if len(self.__pairs) == self.__max_unique_pairs_expected:
            # no reasons to search further - all pairs are found
            raise StopIteration()

        previous_unique_pairs_count = len(self.__pairs)
        chosen_values_arr = [None] * len(self.__working_item_matrix)
        indexes = [None] * len(self.__working_item_matrix)

        direction = 1
        i = 0

        while -1 < i < len(self.__working_item_matrix):
            if direction == 1:
                # move forward
                self.__resort_working_array(chosen_values_arr[:i], i)
                indexes[i] = 0
            elif direction == 0 or direction == -1:
                # scan current array or go back
                indexes[i] += 1
                if indexes[i] >= len(self.__working_item_matrix[i]):
                    direction = -1
                    if i == 0:
                        raise StopIteration()
                    i += direction
                    continue
                direction = 0
            else:
                raise ValueError(
                    "next(): unknown 'direction' code '{}'".format(direction))

            chosen_values_arr[i] = self.__working_item_matrix[i][indexes[i]]

            if self.__filter_func(
                    self.__get_values_array(chosen_values_arr[:i + 1])):
                assert(direction > -1)
                direction = 1
            else:
                direction = 0
            i += direction

        if len(self.__working_item_matrix) != len(chosen_values_arr):
            raise StopIteration()

        self.__pairs.add_sequence(chosen_values_arr)

        if len(self.__pairs) == previous_unique_pairs_count:
            # could not find new unique pairs - stop
            raise StopIteration()

        # replace returned array elements with real values and return it
        return self.__get_iteration_value(chosen_values_arr)

    def __validate_parameter(self, value):
        if isinstance(value, OrderedDict):
            for parameter_list in six.itervalues(value):
                if not parameter_list:
                    raise ValueError(
                        "each parameter arrays must have at least one item")

            return

        if len(value) < 2:
            raise ValueError("must provide more than one option")

        for parameter_list in value:
            if not parameter_list:
                raise ValueError(
                    "each parameter arrays must have at least one item")

    def __resort_working_array(self, chosen_values_arr, num):
        for item in self.__working_item_matrix[num]:
            data_node = self.__pairs.get_node_info(item)

            new_combs = []
            for i in range(0, self.__n):
                # numbers of new combinations to be created if this item is
                # appended to array
                new_combs.append(set([
                    key(z) for z in combinations(
                        chosen_values_arr + [item], i + 1)
                ]) - self.__pairs.get_combs()[i])

            # weighting the node
            # node that creates most of new pairs is the best
            item.weights = [-len(new_combs[-1])]

            # less used outbound connections most likely to produce more new
            # pairs while search continues
            item.weights += [len(data_node.out)]
            item.weights += [len(x) for x in reversed(new_combs[:-1])]
            item.weights += [-data_node.counter]  # less used node is better

            # otherwise we will prefer node with most of free inbound
            # connections; somehow it works out better ;)
            item.weights += [-len(data_node.in_)]

        self.__working_item_matrix[num].sort(key=cmp_to_key(cmp_item))

    def __get_working_item_matrix(self, parameter_matrix):
        return [
            [
                Item("a{:d}v{:d}".format(param_idx, value_idx), value)
                for value_idx, value in enumerate(value_list)
            ]
            for param_idx, value_list in enumerate(parameter_matrix)
        ]

    def __get_values_array(self, item_list):
        return [item.value for item in item_list]

    def __get_iteration_value(self, item_list):
        if not self.__param_name_list:
            return [item.value for item in item_list]

        return self.__Pairs(*[item.value for item in item_list])

    def __extract_param_name_list(self, parameters):
        if not self.__is_ordered_dict_param:
            return []

        return list(parameters)

    def __extract_value_matrix(self, parameters):
        if not self.__is_ordered_dict_param:
            return parameters

        return [v for v in (six.itervalues(parameters))]



# encoding: utf-8

"""
.. codeauthor:: Tsuyoshi Hombashi <gogogo.vm@gmail.com>
"""

from collections import OrderedDict

import pytest

from allpairspy import AllPairs


class Test_pairewise_OrderedDict(object):

    def test_normal(self):
        parameters = OrderedDict({
            "brand": ["Brand X", "Brand Y"],
            "os": ["NT", "2000", "XP"],
            "minute": [15, 30, 60],
        })

        for pairs in AllPairs(parameters):
            assert pairs.brand == 'Brand X'
            assert pairs.os == 'NT'
            assert pairs.minute == 15
            break

        assert len(list(AllPairs(parameters))) == 9


class Test_pairewise_list(object):
    # example1.1.py

    def test_normal(self):
        parameters = [
            ["Brand X", "Brand Y"],
            ["98", "NT", "2000", "XP"],
            ["Internal", "Modem"],
            ["Salaried", "Hourly", "Part-Time", "Contr."],
            [6, 10, 15, 30, 60],
        ]

        assert list(AllPairs(parameters)) == [
            ['Brand X', '98', 'Internal', 'Salaried', 6],
            ['Brand Y', 'NT', 'Modem', 'Hourly', 6],
            ['Brand Y', '2000', 'Internal', 'Part-Time', 10],
            ['Brand X', 'XP', 'Modem', 'Contr.', 10],
            ['Brand X', '2000', 'Modem', 'Part-Time', 15],
            ['Brand Y', 'XP', 'Internal', 'Hourly', 15],
            ['Brand Y', '98', 'Modem', 'Salaried', 30],
            ['Brand X', 'NT', 'Internal', 'Contr.', 30],
            ['Brand X', '98', 'Internal', 'Hourly', 60],
            ['Brand Y', '2000', 'Modem', 'Contr.', 60],
            ['Brand Y', 'NT', 'Modem', 'Salaried', 60],
            ['Brand Y', 'XP', 'Modem', 'Part-Time', 60],
            ['Brand Y', '2000', 'Modem', 'Hourly', 30],
            ['Brand Y', '98', 'Modem', 'Contr.', 15],
            ['Brand Y', 'XP', 'Modem', 'Salaried', 15],
            ['Brand Y', 'NT', 'Modem', 'Part-Time', 15],
            ['Brand Y', 'XP', 'Modem', 'Part-Time', 30],
            ['Brand Y', '98', 'Modem', 'Part-Time', 6],
            ['Brand Y', '2000', 'Modem', 'Salaried', 6],
            ['Brand Y', '98', 'Modem', 'Salaried', 10],
            ['Brand Y', 'XP', 'Modem', 'Contr.', 6],
            ['Brand Y', 'NT', 'Modem', 'Hourly', 10]
        ]


class Test_triplewise(object):
    # example1.2.py

    def test_normal(self):
        parameters = [
            ["Brand X", "Brand Y"],
            ["98", "NT", "2000", "XP"],
            ["Internal", "Modem"],
            ["Salaried", "Hourly", "Part-Time", "Contr."],
            [6, 10, 15, 30, 60],
        ]

        assert list(AllPairs(parameters, n=3)) == [
            ['Brand X', '98', 'Internal', 'Salaried', 6],
            ['Brand Y', 'NT', 'Modem', 'Hourly', 6],
            ['Brand Y', '2000', 'Modem', 'Part-Time', 10],
            ['Brand X', 'XP', 'Internal', 'Contr.', 10],
            ['Brand X', 'XP', 'Modem', 'Part-Time', 6],
            ['Brand Y', '2000', 'Internal', 'Hourly', 15],
            ['Brand Y', 'NT', 'Internal', 'Salaried', 10],
            ['Brand X', '98', 'Modem', 'Contr.', 15],
            ['Brand X', '98', 'Modem', 'Hourly', 10],
            ['Brand Y', 'NT', 'Modem', 'Contr.', 30],
            ['Brand X', 'XP', 'Internal', 'Hourly', 30],
            ['Brand X', '2000', 'Modem', 'Salaried', 30],
            ['Brand Y', '2000', 'Internal', 'Contr.', 6],
            ['Brand Y', 'NT', 'Internal', 'Part-Time', 60],
            ['Brand Y', 'XP', 'Modem', 'Salaried', 15],
            ['Brand X', '98', 'Modem', 'Part-Time', 60],
            ['Brand X', 'XP', 'Modem', 'Salaried', 60],
            ['Brand X', '2000', 'Internal', 'Part-Time', 15],
            ['Brand X', '2000', 'Modem', 'Contr.', 60],
            ['Brand X', '98', 'Modem', 'Salaried', 10],
            ['Brand X', '98', 'Modem', 'Part-Time', 30],
            ['Brand X', 'NT', 'Modem', 'Part-Time', 10],
            ['Brand Y', 'NT', 'Modem', 'Salaried', 60],
            ['Brand Y', 'NT', 'Modem', 'Hourly', 15],
            ['Brand Y', 'NT', 'Modem', 'Hourly', 30],
            ['Brand Y', 'NT', 'Modem', 'Hourly', 60],
            ['Brand Y', 'NT', 'Modem', 'Hourly', 10],
        ]


class Test_pairewise_w_tested(object):
    # example1.3.py

    def test_normal(self):
        parameters = [
            ["Brand X", "Brand Y"],
            ["98", "NT", "2000", "XP"],
            ["Internal", "Modem"],
            ["Salaried", "Hourly", "Part-Time", "Contr."],
            [6, 10, 15, 30, 60],
        ]
        tested = [
            ["Brand X", "98", "Modem", "Hourly", 10],
            ["Brand X", "98", "Modem", "Hourly", 15],
            ["Brand Y", "NT", "Internal", "Part-Time", 10],
        ]

        assert list(AllPairs(parameters, previously_tested=tested)) == [
            ['Brand Y', '2000', 'Modem', 'Salaried', 6],
            ['Brand X', 'XP', 'Internal', 'Contr.', 6],
            ['Brand Y', 'XP', 'Modem', 'Contr.', 30],
            ['Brand X', '2000', 'Internal', 'Part-Time', 30],
            ['Brand Y', '98', 'Internal', 'Salaried', 60],
            ['Brand X', 'NT', 'Modem', 'Salaried', 60],
            ['Brand Y', 'XP', 'Internal', 'Hourly', 15],
            ['Brand Y', 'NT', 'Modem', 'Hourly', 30],
            ['Brand Y', '2000', 'Modem', 'Part-Time', 15],
            ['Brand Y', '2000', 'Modem', 'Contr.', 10],
            ['Brand Y', 'XP', 'Modem', 'Salaried', 10],
            ['Brand Y', '98', 'Modem', 'Part-Time', 6],
            ['Brand Y', 'NT', 'Modem', 'Contr.', 15],
            ['Brand Y', '98', 'Modem', 'Contr.', 30],
            ['Brand Y', 'XP', 'Modem', 'Part-Time', 60],
            ['Brand Y', '2000', 'Modem', 'Hourly', 60],
            ['Brand Y', 'NT', 'Modem', 'Salaried', 30],
            ['Brand Y', 'NT', 'Modem', 'Salaried', 15],
            ['Brand Y', 'NT', 'Modem', 'Hourly', 6],
            ['Brand Y', 'NT', 'Modem', 'Contr.', 60],
        ]


class Test_pairewise_filter(object):

    def test_normal_example21(self):
        # example2.1.py

        parameters = [
            ["Brand X", "Brand Y"],
            ["98", "NT", "2000", "XP"],
            ["Internal", "Modem"],
            ["Salaried", "Hourly", "Part-Time", "Contr."],
            [6, 10, 15, 30, 60]
        ]

        def is_valid_combination(row):
            """
            Should return True if combination is valid and False otherwise.

            Test row that is passed here can be incomplete.
            To prevent search for unnecessary items filtering function
            is executed with found subset of data to validate it.
            """

            n = len(row)
            if n > 1:
                # Brand Y does not support Windows 98
                if "98" == row[1] and "Brand Y" == row[0]:
                    return False
                # Brand X does not work with XP
                if "XP" == row[1] and "Brand X" == row[0]:
                    return False
            if n > 4:
                # Contractors are billed in 30 min increments
                if "Contr." == row[3] and row[4] < 30:
                    return False

            return True

        assert list(AllPairs(
            parameters, filter_func=is_valid_combination)
        ) == [
            ['Brand X', '98', 'Internal', 'Salaried', 6],
            ['Brand Y', 'NT', 'Modem', 'Hourly', 6],
            ['Brand Y', '2000', 'Internal', 'Part-Time', 10],
            ['Brand X', '2000', 'Modem', 'Contr.', 30],
            ['Brand X', 'NT', 'Internal', 'Contr.', 60],
            ['Brand Y', 'XP', 'Modem', 'Salaried', 60],
            ['Brand X', '98', 'Modem', 'Part-Time', 15],
            ['Brand Y', 'XP', 'Internal', 'Hourly', 15],
            ['Brand Y', 'NT', 'Internal', 'Part-Time', 30],
            ['Brand X', '2000', 'Modem', 'Hourly', 10],
            ['Brand Y', 'XP', 'Modem', 'Contr.', 30],
            ['Brand Y', '2000', 'Modem', 'Salaried', 15],
            ['Brand Y', 'NT', 'Modem', 'Salaried', 10],
            ['Brand Y', 'XP', 'Modem', 'Part-Time', 6],
            ['Brand Y', '2000', 'Modem', 'Contr.', 60],
        ]

    def test_normal_example22(self):
        # example2.2.py

        parameters = [
            ("brand", ["Brand X", "Brand Y"]),
            ("os", ["98", "NT", "2000", "XP"]),
            ("network", ["Internal", "Modem"]),
            ("employee", ["Salaried", "Hourly", "Part-Time", "Contr."]),
            ("increment", [6, 10, 15, 30, 60])
        ]

        def is_valid_combination(values, names):

            dictionary = dict(zip(names, values))

            """
            Should return True if combination is valid and False otherwise.

            Dictionary that is passed here can be incomplete.
            To prevent search for unnecessary items filtering function
            is executed with found subset of data to validate it.
            """

            rules = [
                # Brand Y does not support Windows 98
                # Brand X does not work with XP
                # Contractors are billed in 30 min increments
                lambda d: "98" == d["os"] and "Brand Y" == d["brand"],
                lambda d: "XP" == d["os"] and "Brand X" == d["brand"],
                lambda d: "Contr." == d["employee"] and d["increment"] < 30
            ]

            for rule in rules:
                try:
                    if rule(dictionary):
                        return False
                except KeyError:
                    pass

            return True

        assert list(AllPairs(
            [x[1] for x in parameters],
            filter_func=lambda values: is_valid_combination(
                values, [x[0] for x in parameters])
        )) == [
            ['Brand X', '98', 'Internal', 'Salaried', 6],
            ['Brand Y', 'NT', 'Modem', 'Hourly', 6],
            ['Brand Y', '2000', 'Internal', 'Part-Time', 10],
            ['Brand X', '2000', 'Modem', 'Contr.', 30],
            ['Brand X', 'NT', 'Internal', 'Contr.', 60],
            ['Brand Y', 'XP', 'Modem', 'Salaried', 60],
            ['Brand X', '98', 'Modem', 'Part-Time', 15],
            ['Brand Y', 'XP', 'Internal', 'Hourly', 15],
            ['Brand Y', 'NT', 'Internal', 'Part-Time', 30],
            ['Brand X', '2000', 'Modem', 'Hourly', 10],
            ['Brand Y', 'XP', 'Modem', 'Contr.', 30],
            ['Brand Y', '2000', 'Modem', 'Salaried', 15],
            ['Brand Y', 'NT', 'Modem', 'Salaried', 10],
            ['Brand Y', 'XP', 'Modem', 'Part-Time', 6],
            ['Brand Y', '2000', 'Modem', 'Contr.', 60],
        ]



# encoding: utf-8

from itertools import combinations

from six.moves import range


class Node(object):

    @property
    def id(self):
        return self.__node_id

    @property
    def counter(self):
        return self.__counter

    def __init__(self, node_id):
        self.__node_id = node_id
        self.__counter = 0
        self.in_ = set()
        self.out = set()

    def __str__(self):
        return str(self.__dict__)

    def inc_counter(self):
        self.__counter += 1


def key(items):
    return "->".join([x.id for x in items])


class PairsStorage(object):

    def __init__(self, n):
        self.__n = n
        self.__nodes = {}
        self.__combs_arr = [
            set() for _i in range(n)
        ]

    def __len__(self):
        return len(self.__combs_arr[-1])

    def add_sequence(self, sequence):
        for i in range(1, self.__n + 1):
            for combination in combinations(sequence, i):
                self.__add_combination(combination)

    def get_node_info(self, item):
        return self.__nodes.get(item.id, Node(item.id))

    def get_combs(self):
        return self.__combs_arr

    def __add_combination(self, combination):
        n = len(combination)
        assert(n > 0)

        self.__combs_arr[n - 1].add(key(combination))
        if n == 1 and combination[0].id not in self.__nodes:
            self.__nodes[combination[0].id] = Node(combination[0].id)
            return

        ids = [x.id for x in combination]
        for i, id in enumerate(ids):
            curr = self.__nodes[id]
            curr.inc_counter()
            curr.in_.update(ids[:i])
            curr.out.update(ids[i + 1:])



# encoding: utf-8

from distutils.core import setup
import os.path
import sys

import setuptools


REQUIREMENT_DIR = "requirements"

needs_pytest = set(["pytest", "test", "ptr"]).intersection(sys.argv)
pytest_runner = ["pytest-runner"] if needs_pytest else []


with open(os.path.join(REQUIREMENT_DIR, "requirements.txt")) as f:
    install_requires = [line.strip() for line in f if line.strip()]

with open(os.path.join(REQUIREMENT_DIR, "test_requirements.txt")) as f:
    tests_require = [line.strip() for line in f if line.strip()]

setup(
    name="allpairspy",
    version="2.3.0",
    description="Pairwise test combinations generator",
    long_description="""Pairwise (aka 'all-pairs') test combinations generator written in
Python. Allows one to create a set of tests using 'pairwise combinations' method,
reducing a number of combinations of variables
into a lesser set that covers most situations.
""",
    author="Tsuyoshi Hombashi",
    author_email="gogogo.vm@gmail.com",
    url="https://github.com/thombashi/allpairspy",
    install_requires=install_requires,
    tests_require=tests_require,
    setup_requires=pytest_runner,
    packages=setuptools.find_packages(exclude=["test*"]),
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Environment :: Console",
        "Intended Audience :: Developers",
        "Intended Audience :: Information Technology",
        "Intended Audience :: System Administrators",
        "License :: OSI Approved :: MIT License",
        "Natural Language :: English",
        "Operating System :: OS Independent",
        "Programming Language :: Python",
        "Programming Language :: Python :: 2",
        "Programming Language :: Python :: 2.7",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.3",
        "Programming Language :: Python :: 3.4",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Topic :: Software Development :: Libraries",
        "Topic :: Software Development :: Libraries :: Python Modules",
        "Topic :: Software Development :: Testing",
        "Topic :: Utilities",
    ]
)
